#define MAX_FACES 2000
#define MAX_VERTICES MAX_FACES*3
#define CROSS_SECTION_VERTS 8
#define PI 3.14159265358979323846

#include <fstream>
#include <math.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include "image.h"

//std::ofstream outfile("stdout.txt");
int ImageLoadFromResource(LPCTSTR lpszName, Image *image) {
	HANDLE hBmp = LoadImage(hInstance, lpszName, IMAGE_BITMAP,
							LR_DEFAULTSIZE, LR_DEFAULTSIZE, LR_DEFAULTCOLOR);
	if(!hBmp)
		return 0;

	// Get the bitmap dimensions.
	BITMAP bmp;
	if(!GetObject((HBITMAP)hBmp, sizeof(bmp), &bmp))
		return 0;

	// First try to get how much memory we need.
	BITMAPINFO bi = {0};
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biWidth = bmp.bmWidth;
	bi.bmiHeader.biHeight = bmp.bmHeight;
	bi.bmiHeader.biPlanes = bmp.bmPlanes;
	bi.bmiHeader.biBitCount = 24;
	if(!GetDIBits(hDC, (HBITMAP)hBmp, 0, 1, NULL, &bi, DIB_PAL_COLORS))
		return 0;

	// Allocate memory for the required number of bytes. 
	image->sizeX = bi.bmiHeader.biWidth;
	image->sizeY = bi.bmiHeader.biHeight;
	image->data = (char*)malloc(bi.bmiHeader.biSizeImage);

	// Get the bits in 24 bit format.
	if(!GetDIBits(hDC, (HBITMAP)hBmp, 0, bi.bmiHeader.biHeight, (LPVOID)image->data, &bi, DIB_PAL_COLORS))
		return 0;

    for (unsigned long i=0;i<bi.bmiHeader.biSizeImage;i+=3) { /* reverse all of the colors. (bgr -> rgb)*/
	char temp = image->data[i];
	image->data[i] = image->data[i+2];
	image->data[i+2] = temp;
    }

	return 1;
}

/*************************************************************************************/
/*************************************************************************************/
/***        Load Bitmaps And Convert To Textures                                  ****/
/*************************************************************************************/
int LoadGLTextures(int rcID) {
    /* Load Texture*/
    Image *image1;
    static GLuint	texture[1];
    
    /* allocate space for texture*/
    image1 = (Image *) malloc(sizeof(Image));
    if (image1 == NULL) {
	printf("Error allocating space for image");
	exit(0);
    }

    if (!ImageLoadFromResource(MAKEINTRESOURCE(rcID), image1)) {
	exit(1);
    } 

    /* Create Texture	*****************************************/
    glGenTextures(1, &texture[0]);
    glBindTexture(GL_TEXTURE_2D, texture[0]);   /* 2d texture (x and y size)*/

    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR); /* scale linearly when image bigger than texture*/
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); /* scale linearly when image smalled than texture*/
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);

    /* 2d texture, level of detail 0 (normal), 3 components (red, green, blue), x size from image, y size from image, */
    /* border 0 (normal), rgb color data, unsigned byte data, and finally the data itself.*/
    glTexImage2D(GL_TEXTURE_2D, 0, 3, image1->sizeX, image1->sizeY, 0, GL_RGB, GL_UNSIGNED_BYTE, image1->data);

	free(image1->data);
	free(image1);

    return texture[0];
}

struct VERTEX {
       float x,y,z;
       float tx, ty; // texture coords
       float nx, ny, nz;
       int nFaces;
};

struct INDEX {
       int nVertex;
};

struct FACE {
       INDEX pIndices[3];
       int nCount;
       VERTEX vNormal;
};

float calc_distance(VERTEX v1, VERTEX v2) {
   float x = v1.x-v2.x;
   float y = v1.y-v2.y;
   float z = v1.z-v2.z;
   return sqrt(x*x+y*y+z*z);
}

float calc_length(VERTEX v) {
   return sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
}

float calc_length(float vx, float vy, float vz) {
   return sqrt(vx*vx+vy*vy+vz*vz);
}

class CPolygon
{
      private:
              VERTEX m_pVertices[MAX_VERTICES];
              VERTEX m_pOriginalVertices[MAX_VERTICES];
              int m_nVertices;
              FACE m_pFaces[MAX_FACES];
              int m_nFaces;
              int m_iTextureID;
              float m_minx, m_maxx, m_miny, m_maxy, m_minz, m_maxz;

      public:
             CPolygon() : m_nFaces(0), m_nVertices(0), m_minx(0.0),
                          m_maxx(0.0), m_miny(0.0), m_maxy(0.0),
                          m_minz(0.0), m_maxz(0.0), m_iTextureID(0) {}
             ~CPolygon() {static GLuint texName[1] = {m_iTextureID};
			              glDeleteTextures(1, texName);}

             void draw();
             void mapTexture(float size, float mintx, float maxtx, float minty, float maxty);
             int addVertex(float x, float y, float z);
             int addFace(INDEX* pIndex);
             void calcNormal(int nFace);
             void renormalize();

             void save(char* fName);
             void load(char* fName);
             void loadEllipse(char* fName);
             void loadEllipse(float data[][7], int size) ;

             int getTextureID();
             void setTextureID(int texID);
             void loadTextureFile(int rcID);

             float getMinX();
             float getMaxX();
             float getMinY();
             float getMaxY();
             float getMaxZ();
             float getMinZ();

             int getVertexCount();
             VERTEX getVertex(int n);
             VERTEX getOriginalVertex(int n);
             void setVertex(int n, float x, float y, float z);
             void setVertexTexture(int n, float tx, float ty);

             void randomizeVertices(float dx, float dy, float dz);

             void debug();


};

void CPolygon::draw()
{
    //debug();
    //outfile << "Entering draw\n";
    if (m_iTextureID > 0) {
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, m_iTextureID);
    }
    glBegin(GL_TRIANGLES);
    for (int i = 0; i < m_nFaces; i++)
    {
        FACE face = m_pFaces[i];

        for (int f=0; f<face.nCount; f++)
        {
            float x =  m_pVertices[face.pIndices[f].nVertex].x;
            float y =  m_pVertices[face.pIndices[f].nVertex].y;
            float z =  m_pVertices[face.pIndices[f].nVertex].z;
            float nx =  m_pVertices[face.pIndices[f].nVertex].nx;
            float ny =  m_pVertices[face.pIndices[f].nVertex].ny;
            float nz =  m_pVertices[face.pIndices[f].nVertex].nz;
            float tx = m_pVertices[face.pIndices[f].nVertex].tx;
            float ty = m_pVertices[face.pIndices[f].nVertex].ty;
            glTexCoord2f( tx , ty );
            glNormal3f( nx, ny, nz);
            glVertex3f( x, y, z );
        }
    }    
    glEnd();
    if (m_iTextureID > 0) {
        glDisable(GL_TEXTURE_2D);
    }
}

void CPolygon::mapTexture(float size, float mintx, float maxtx, float minty, float maxty) {
    for (int i = 0; i < m_nVertices; i++) {
        m_pVertices[i].tx = ((maxtx+1)/size)*(m_pVertices[i].x - m_minx)/(m_maxx-m_minx) + mintx/size;
        m_pVertices[i].ty = ((maxty+1)/size)*(m_pVertices[i].y - m_miny)/(m_maxy-m_miny) + minty/size;
        //outfile << m_pVertices[i].x << " " << m_pVertices[i].y << " " << m_pVertices[i].tx << " " << m_pVertices[i].ty << "\n" ;
    }
}

int CPolygon::addVertex(float x, float y, float z) {
    //outfile << "Entering addVertex("<<x<<","<<y<<","<<z<<")\n";
    if (m_nVertices == MAX_VERTICES) {
      return -1;
    }
    // check if we already have this vertex
    for (int i = 0; i < m_nVertices; i++) {
        if (m_pVertices[i].x == x &&
            m_pVertices[i].y == y &&
            m_pVertices[i].z == z) {
            return i;
        }
    }
    if (x > m_maxx) m_maxx = x;
    if (x < m_minx) m_minx = x;
    if (y > m_maxy) m_maxy = y;
    if (y < m_miny) m_miny = y;
    if (z > m_maxz) m_maxz = z;
    if (z < m_minz) m_minz = z;

    // if not, return a new one
    m_pVertices[m_nVertices].x = x;
    m_pVertices[m_nVertices].y = y;
    m_pVertices[m_nVertices].z = z;
    m_pVertices[m_nVertices].nx = 0.0;
    m_pVertices[m_nVertices].ny = 0.0;
    m_pVertices[m_nVertices].nz = 0.0;
    m_pVertices[m_nVertices].nFaces = 0;

    return m_nVertices++;
}

int CPolygon::addFace(INDEX* pIndices) {
    //outfile << "Entering addFace\n";
    if (m_nFaces == MAX_FACES) {
      return -1;
    }

    if (pIndices[0].nVertex < 0 || pIndices[1].nVertex  < 0 || pIndices[2].nVertex  < 0) {
       return -1;
    }

    m_pFaces[m_nFaces].pIndices[0] = pIndices[0];
    m_pFaces[m_nFaces].pIndices[1] = pIndices[1];
    m_pFaces[m_nFaces].pIndices[2] = pIndices[2];

    m_pFaces[m_nFaces].nCount = 3;

    calcNormal(m_nFaces);

    return m_nFaces++;
}
/*
void CPolygon::debug() {
     outfile << "--------------------\n";
     outfile << "m_nFaces = " << m_nFaces << "\n";
     for (int i = 0; i < m_nFaces; i++) {
         outfile << "face " << i << " nCount = " << m_pFaces[i].nCount << "\n";
         for (int j = 0; j < m_pFaces[i].nCount; j++) {
             outfile << "face " << i << " vertex " << j << " = " << m_pFaces[i].pIndices[j].nVertex << "\n";
         }
     }
     outfile << "\nm_nVertices = " << m_nVertices << "\n";
     for (int i = 0; i < m_nVertices; i++) {
         outfile << "vertex " << i << " = (" << m_pVertices[i].x << "," << m_pVertices[i].y << "," << m_pVertices[i].z << ")\n";
     }
}
 */
void CPolygon::calcNormal(int nFace) {
    GLfloat v1[3];
    GLfloat v2[3];
    GLfloat norm[3];
    GLfloat divisor;

    FACE f = m_pFaces[nFace];

    if(f.nCount == 3) {
        v1[0] = m_pVertices[f.pIndices[0].nVertex].x - m_pVertices[f.pIndices[1].nVertex].x;
        v1[1] = m_pVertices[f.pIndices[0].nVertex].y - m_pVertices[f.pIndices[1].nVertex].y;
        v1[2] = m_pVertices[f.pIndices[0].nVertex].z - m_pVertices[f.pIndices[1].nVertex].z;

        v2[0] = m_pVertices[f.pIndices[1].nVertex].x - m_pVertices[f.pIndices[2].nVertex].x;
        v2[1] = m_pVertices[f.pIndices[1].nVertex].y - m_pVertices[f.pIndices[2].nVertex].y;
        v2[2] = m_pVertices[f.pIndices[1].nVertex].z - m_pVertices[f.pIndices[2].nVertex].z;

        // cross product
        norm[0] = v1[1]*v2[2] - v1[2]*v2[1];
        norm[1] = v1[0]*v2[2] - v1[2]*v2[0];
        norm[2] = v1[0]*v2[1] - v1[1]*v2[0];

        // make it unit length
        divisor = norm[0]*norm[0] + norm[1]*norm[1] + norm[2]*norm[2];
        if (divisor != 0) {
           divisor = sqrt(divisor);
           norm[0] /= divisor;
           norm[1] /= divisor;
           norm[2] /= divisor;
        }

        m_pFaces[nFace].vNormal.x = norm[0];
        m_pFaces[nFace].vNormal.y = norm[1];
        m_pFaces[nFace].vNormal.z = norm[2];

        for (int i = 0 ; i < 3; i++) {
            m_pVertices[f.pIndices[i].nVertex].nx =
                      (m_pVertices[f.pIndices[i].nVertex].nx*m_pVertices[f.pIndices[i].nVertex].nFaces +
                      m_pFaces[nFace].vNormal.x)/(m_pVertices[f.pIndices[i].nVertex].nFaces + 1);
            m_pVertices[f.pIndices[i].nVertex].ny =
                      (m_pVertices[f.pIndices[i].nVertex].ny*m_pVertices[f.pIndices[i].nVertex].nFaces +
                      m_pFaces[nFace].vNormal.y)/(m_pVertices[f.pIndices[i].nVertex].nFaces + 1);
            m_pVertices[f.pIndices[i].nVertex].nz =
                      (m_pVertices[f.pIndices[i].nVertex].nz*m_pVertices[f.pIndices[i].nVertex].nFaces +
                      m_pFaces[nFace].vNormal.z)/(m_pVertices[f.pIndices[i].nVertex].nFaces + 1);
            m_pVertices[f.pIndices[i].nVertex].nFaces++;
        }
    }


}

void CPolygon::save(char* fName) {
    std::ofstream sv(fName);

    for (int i = 0; i < m_nFaces; i++) {
        for (int j = 0; j < m_pFaces[i].nCount; j++) {
            int n = m_pFaces[i].pIndices[j].nVertex;
            sv << m_pVertices[n].x << " " << m_pVertices[n].y << " " << m_pVertices[n].z << "\n";
        }
        sv << "\n";
    }
    sv.close();
}

void CPolygon::load(char* fName) {
    std::ifstream ld(fName);
    char ch[256], *pch;
    INDEX verts[3];
    int vCount = 0, n = 0;
    float vals[3];
    while (ld.getline(ch,256)) {
        n = 0;
        if (strlen(ch) > 0) {
           pch = strtok (ch, ",");
           while (pch != NULL)
           {
               vals[n] = atof(pch);
               pch = strtok (NULL, ",");

               n++;
           }
           verts[vCount].nVertex = addVertex(vals[0], vals[1], vals[2]);
           vCount++;
        }
        else if (vCount == 3) {
           vCount = 0;
           addFace(verts);
        }
        else {
           // do jack
        }
    }
}

void CPolygon::loadEllipse(float data[][7], int datasize) {
    INDEX vertsIdx[3];
    int nVerts = CROSS_SECTION_VERTS, n = 0, vCount = 0;
    int nLines = 0;
    float dx, dy, dz, w, h;

    float x, y, z, angle, dangle;
    float size, mintx, maxtx, minty, maxty;
    int count = 0, rcID;
    for (int i = 0; i < datasize; i++ ) {
        n = 0;

        if (i == 0) {
           rcID = (int)data[i][0];
        }
        else if (i == 1) {
           size = data[i][0];
           mintx = data[i][1];
           maxtx = data[i][2];
           minty = data[i][3];
           maxty = data[i][4];
        }
        else if (count == 2) {
             // Do nothing, behavior stuff here
        }
        else {
           dx = data[i][0];
           dy = data[i][1];
           dz = data[i][2];
           w = data[i][3];
           h = data[i][4];
           angle = 0.0;
           dangle = 2 * PI / nVerts;
           for (int i = 0; i < nVerts; i++) {
               x = dx;
               y = h * cos(angle) + dy;
               z = w * sin(angle) + dz;
               addVertex(x, y, z);
               //outfile <<n<<" "<<x<<" "<<y<<" "<<z<<"\n";
               angle += dangle;
           }
           nLines++;

        }
        count++;
    }

    for (i = 0; i < (nLines) * (nVerts) - (nVerts+2); i++) {
        vertsIdx[0].nVertex = i;
        vertsIdx[1].nVertex = i+1;
        vertsIdx[2].nVertex = i+nVerts;
        addFace(vertsIdx);
        vertsIdx[0].nVertex = i+1;
        vertsIdx[1].nVertex = i+1+nVerts;
        vertsIdx[2].nVertex = i+nVerts;
        addFace(vertsIdx);
    }

    for (i = 0; i < m_nVertices; i++) {
       m_pOriginalVertices[i] = m_pVertices[i];
    }

    // Load the texture data

    if (m_iTextureID == 0) m_iTextureID = LoadGLTextures(rcID);
    mapTexture(size, mintx, maxtx, minty, maxty);
}

void CPolygon::loadEllipse(char* fName) {
    std::ifstream ld(fName);
    char ch[256], *pch;
    INDEX vertsIdx[3];
    int nVerts = CROSS_SECTION_VERTS, n = 0, vCount = 0;
    int nLines = 0;
    float vals[5], dx, dy, dz, w, h;

    float x, y, z, angle, dangle;
    float size, mintx, maxtx, minty, maxty;
    int count = 0, rcID;
    while (ld.getline(ch,256)) {
        n = 0;

        if (count == 0) {
           rcID = atoi(ch);

        }
        else if (count == 1) {
           pch = strtok (ch, ",");
           while (pch != NULL)
           {
               vals[n] = atof(pch);
               pch = strtok (NULL, ",");

               n++;
           }
           size = vals[0];
           mintx = vals[1];
           maxtx = vals[2];
           minty = vals[3];
           maxty = vals[4];
        }
        else if (count == 2) {
             // Do nothing, behavior stuff here
        }
        else if (strlen(ch) > 0) {
           pch = strtok (ch, ",");
           while (pch != NULL)
           {
               vals[n] = atof(pch);
               pch = strtok (NULL, ",");

               n++;
           }
           dx = vals[0];
           dy = vals[1];
           dz = vals[2];
           w = vals[3];
           h = vals[4];
           angle = 0.0;
           dangle = 2 * PI / nVerts;
           for (int i = 0; i < nVerts; i++) {
               x = dx;
               y = h * cos(angle) + dy;
               z = w * sin(angle) + dz;
               addVertex(x, y, z);
               //outfile <<n<<" "<<x<<" "<<y<<" "<<z<<"\n";
               angle += dangle;
           }
           nLines++;

        }
        count++;
    }
    ld.close();
    for (int i = 0; i < (nLines) * (nVerts) - (nVerts+2); i++) {
        vertsIdx[0].nVertex = i;
        vertsIdx[1].nVertex = i+1;
        vertsIdx[2].nVertex = i+nVerts;
        addFace(vertsIdx);
        vertsIdx[0].nVertex = i+1;
        vertsIdx[1].nVertex = i+1+nVerts;
        vertsIdx[2].nVertex = i+nVerts;
        addFace(vertsIdx);
    }

    for (i = 0; i < m_nVertices; i++) {
       m_pOriginalVertices[i] = m_pVertices[i];
    }

    // Load the texture data
    //outfile << rcID;
    if (m_iTextureID == 0) m_iTextureID = LoadGLTextures(rcID);
    mapTexture(size, mintx, maxtx, minty, maxty);
}

int CPolygon::getTextureID() {
    return m_iTextureID;
}

void CPolygon::setTextureID(int texID) {
    m_iTextureID = texID;
}

float CPolygon::getMinX() {
   return m_minx;
}

float CPolygon::getMaxX() {
   return m_maxx;
}

float CPolygon::getMinY() {
   return m_miny;
}

float CPolygon::getMaxY() {
   return m_maxy;
}

float CPolygon::getMinZ() {
   return m_minz;
}

float CPolygon::getMaxZ() {
   return m_maxz;
}

int CPolygon::getVertexCount() {
    return m_nVertices;
}

VERTEX CPolygon::getVertex(int n) {
    if (n < m_nVertices) {
        return m_pVertices[n];
    }
    else {
        VERTEX v;
        return v;
    }
}

VERTEX CPolygon::getOriginalVertex(int n) {
    if (n < m_nVertices) {
        return m_pOriginalVertices[n];
    }
    else {
        VERTEX v;
        return v;
    }
}

void CPolygon::setVertex(int n, float x, float y, float z) {
    if (n < m_nVertices) {
        m_pVertices[n].x = x;
        m_pVertices[n].y = y;
        m_pVertices[n].z = z;
    }
}

void CPolygon::setVertexTexture(int n, float tx, float ty) {
    m_pVertices[n].tx = tx;
    m_pVertices[n].ty = ty;
}

void CPolygon::loadTextureFile(int rcID) {
    m_iTextureID = LoadGLTextures(rcID);
}

void CPolygon::randomizeVertices(float dx, float dy, float dz) {
    float rnd = 0.0, rndmax = RAND_MAX;
    for (int i = 0; i < m_nVertices; i++) {
        rnd = rand();
        m_pVertices[i].x += (rnd/rndmax)*2*dx - dx;
        rnd = rand();
        m_pVertices[i].y += (rnd/rndmax)*2*dy - dy;
        rnd = rand();
        m_pVertices[i].z += (rnd/rndmax)*2*dz - dz;
    }
}

void CPolygon::renormalize() {
    for (int i = 0; i < m_nFaces; i++)
    {
        calcNormal(i);
    }    
}
