#include <math.h>
#include <stdio.h>

#include "MySaver.h"

//-----------------------------------------------------------------------------

// Disable spurious VC6 warnings about floating point truncation
#ifdef  _MSC_VER
#pragma warning (disable:4305 4244)
#endif//_MSC_VER

//-----------------------------------------------------------------------------

const float
    DegRad = 3.141592654 / 180.0;	// degrees to radians

// Various constants that control the screen saver effects
//
const float
    V_RASTER_JITTER_AMP  = 0.1,     // vertical jitter amplitude
    V_RASTER_JITTER_RATE = 100.0,   // vertical jitter rate
    V_RASTER_JITTER_ROT  = 0.5,     // strength of rotation jitter
    RASTER_FADE_AMP      = 0.05,    // amplitude of raster fade effect (0..1)
    RASTER_FADE_RATE     = 0.5;     // raster fade cycle rate

// Texture list IDs - would be preferable to use bind texture, but
// that isn't necessarily supported on all cards/drivers.
//
enum {
    TEX_RASTER  = 100,  // raster texture map
    TEX_NUMBER,			// blurry number texture
    TEX_CODE,			// program code texture
    TEX_SNOW,			// snow texture
    TEX_CRTSNAP,		// blurry CRT images
    TEX_KANA			// japanese script
};

// Display list IDs
//
enum {
	LIST_SPHERE = 200	// sphere list base (there are three)
};

//-----------------------------------------------------------------------------

// registry keys & values
#define REG_KEY_APPLICATION "SOFTWARE\\Crema\\RasterSaver"
#define REG_VAL_INSTALL_DIR "InstallDir"

// fallback path (used only if we failed to read from registry - which should
// hopefully never happen unless the installer messed up)
#define FALLBACK_PATH       "C:\\Program Files\\Crema\\RasterSaver"

//-----------------------------------------------------------------------------

/**
 * Constructor
 */
MySaver::MySaver() {
	_bRasters = true;
	_bSnow    = true;
}

//-----------------------------------------------------------------------------

/**
 * Destructor
 */
/*virtual*/ MySaver::~MySaver() {
}

//-----------------------------------------------------------------------------

/**
 * Called when saver starts
 */
void MySaver::OnCreate() {
	//-- get installation path from registry
	
	HKEY appKey;
	bool gotPath = false;

	if (RegOpenKeyEx(
		HKEY_LOCAL_MACHINE, REG_KEY_APPLICATION, 0, KEY_READ, &appKey
	) == ERROR_SUCCESS) {
		// managed to open application registry key

		// read install directory path
		DWORD dwType, dwSize=1024;
		BYTE pBuff[1024];
		if ( (RegQueryValueExA(appKey, REG_VAL_INSTALL_DIR, NULL, &dwType,
			pBuff, &dwSize) == ERROR_SUCCESS) && (dwType == REG_SZ) ) {
			// convert null terminated character string to STL string, and append '\'
			_Path = string((char*)pBuff) + '\\';
			gotPath = true;
		}

		RegCloseKey(appKey);
	}

	// if we failed to read path from registry, assume typical
	// installation directory as fallback
	if (!gotPath) _Path = FALLBACK_PATH;

	// Set timer interval for refresh to 60Hz
	// (may not achieve this on all cards)
	SetTimerInterval(1000/60);

	// Setup OpenGL state
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_CULL_FACE);
}//OnCreate

//-----------------------------------------------------------------------------

// destroy a display list
static void ZapList(GLuint l) {
	// actually, deleting a null list is harmless, but we may as well check
	if (glIsList(l)) glDeleteLists(l,1);
}//ZapList

/**
 * Called when saver stops.
 */
void MySaver::OnDestroy() {
	// clean up display lists
	for (int n=0;n<3;n++)
		ZapList(LIST_SPHERE+n);

	// clean up texture lists
	ZapList(TEX_RASTER);
	ZapList(TEX_NUMBER);
    ZapList(TEX_CODE);
    ZapList(TEX_SNOW);
    ZapList(TEX_CRTSNAP);
    ZapList(TEX_KANA);

	// free models
	//for (n=0;n<4;n++)
	//	_MBuild[n].Free();
}//OnDestroy

//-----------------------------------------------------------------------------

/**
 * Called when OpenGL redraw occurs
 */
void MySaver::OnPaint() {
	// Current time (seconds)
    float Time = (float)Ticks()/1000.0;

    // Width and height of OpenGL window
    int
        width  = Width(),
        height = Height();

    // Ensure height != 0 to avoid divide by zero on w/h aspect calcs.
    if (!height) height++;

    // Keep track of relative time into scene, and switch scenes as required
    static float tStart = -1;
    static int   scene  = 0;
    if (tStart < 0) {
        tStart = Time;
        scene++;
    }
    float TimeRel = Time-tStart;

    // Swap the display buffers
    SwapBuffers();
}//OnPaint

/**
 * Called on timer events
 */
void MySaver::OnTimer() {
	// refresh the display
	Repaint();
}//OnTimer

/**
 * Called when window is resized
 */
void MySaver::OnResize(int width, int height) {
	// ensure OpenGL viewport fills window
	glViewport(0,0,width,height);
}//OnResize

